"""
config_model.py - pydantic definitions for the powermon config model
"""

from typing import Literal, List
from pydantic import BaseModel, Extra, Field


class NoExtraBaseModel(BaseModel):
    """ updated BaseModel with Extras forbidden """
    class Config:
        """pydantic BaseModel config"""
        extra = Extra.forbid


class DaemonConfig(NoExtraBaseModel):
    """ model/allowed elements for daemon section of config """
    type: None | str
    keepalive: None | int


class MQTTConfig(NoExtraBaseModel):
    """ model/allowed elements for mqtt broker section of config """
    name: str
    port: None | int = Field(default=None)
    username: None | str = Field(default=None)
    password: None | str = Field(default=None)


class APIConfig(NoExtraBaseModel):
    """ model/allowed elements for api section of config """
    host: None | str = Field(default=None)
    port: None | int = Field(default=None)
    enabled: None | bool = Field(default=False)
    log_level: None | str = Field(default=None)
    announce_topic: None | str = Field(default=None)
    adhoc_topic: None | str = Field(default=None)
    refresh_interval: None | int = Field(default=None)


class BaseFormatConfig(NoExtraBaseModel):
    """ model/allowed elements for base format config """
    type: str
    tag: None | str = Field(default=None)
    draw_lines: None | bool = Field(default=None)
    keep_case: None | bool = Field(default=None)
    remove_spaces: None | bool = Field(default=None)
    extra_info: None | bool = Field(default=None)
    excl_filter: None | str = Field(default=None)
    filter: None | str = Field(default=None)


class HassFormatConfig(BaseFormatConfig):
    """ model/allowed elements for hass format config """
    discovery_prefix: None | str
    entity_id_prefix: None | str


class MqttFormatConfig(BaseFormatConfig):
    """ model/allowed elements for mqtt format config """
    topic: None | str


class LoopsTriggerConfig(NoExtraBaseModel):
    """ model/allowed elements for 'loops' trigger config """
    loops: int


class AtTriggerConfig(NoExtraBaseModel):
    """ model/allowed elements for 'at' trigger config """
    at: str


class EveryTriggerConfig(NoExtraBaseModel):
    """ model/allowed elements for 'every' trigger config """
    every: int


class OutputConfig(NoExtraBaseModel):
    """ model/allowed elements for output config """
    type: Literal['screen'] | Literal['mqtt'] | Literal['api_mqtt'] | Literal['table']
    format: None | str | BaseFormatConfig | HassFormatConfig | MqttFormatConfig = Field(default=None)


class CommandConfig(NoExtraBaseModel):
    """ model/allowed elements for command section of config """
    command: str
    type: None | Literal["basic"] | Literal["poll"] = Field(default="basic")
    override: None | dict = Field(default=None)
    trigger: None | LoopsTriggerConfig | AtTriggerConfig | EveryTriggerConfig = Field(default=None)
    outputs: None | List[OutputConfig] | str = Field(default=None)


class SerialPortConfig(BaseModel):
    """ model/allowed elements for serial port config """
    type: Literal["serial"]
    path: str
    baud: None | int  = Field(default=None)
    protocol: None | str


class UsbPortConfig(BaseModel):
    """ model/allowed elements for usb port config """
    type: Literal["usb"]
    path: None | str
    protocol: None | str


class TestPortConfig(BaseModel):
    """ model/allowed elements for test port config """
    type: Literal["test"]
    response_number: None | int = Field(default=None)
    protocol: None | str = Field(default=None)


class DeviceConfig(NoExtraBaseModel):
    """ model/allowed elements for device section of config """
    name: None | str = Field(default=None)
    id: None | str | int = Field(default=None)
    model: None | str = Field(default=None)
    manufacturer: None | str = Field(default=None)
    port: TestPortConfig | SerialPortConfig | UsbPortConfig


class BaseConfig(NoExtraBaseModel):
    """ model/allowed elements for first level of config """
    device: DeviceConfig
    commands: List[CommandConfig]
    mqttbroker: None | MQTTConfig = Field(default=None)
    api: None | APIConfig = Field(default=None)
    daemon: None | DaemonConfig = Field(default=None)
    debuglevel: None | int | str = Field(default=None)  # If you put "debug" it translates to 10 then fails to load the config
    loop: None | int | Literal["once"] = Field(default=None)


class ConfigModel(NoExtraBaseModel):
    """Entry point for config model"""
    config: BaseConfig



""" outputs / abstractoutput.py """
import logging
from abc import ABC, abstractmethod

# from powermon.libs.mqttbroker import MqttBroker
from powermon.commands.result import Result
from powermon.dto.outputDTO import OutputDTO
from powermon.formats.abstractformat import AbstractFormat

log = logging.getLogger("Output")


class AbstractOutput(ABC):
    """ base class for all output modules """
    def __init__(self, name=None) -> None:
        self.name = name
        # self.command_code : str = "not_set"
        # self.device_id : str = "not_set"
        self.topic = None

    @property
    def formatter(self):
        """ the formatter for this output """
        return self._formatter

    @formatter.setter
    def formatter(self, formatter : AbstractFormat):
        self._formatter = formatter

    @abstractmethod
    def process(self, command=None, result: Result=None, mqtt_broker=None, device_info=None):
        """ entry point of any output class """
        raise NotImplementedError("need to implement process function")

    # def set_command(self, command_name):
    #     """ set the command_code """
    #     self.command_code = command_name

    # def set_device_id(self, device_id):
    #     """ store the device_id """
    #     self.device_id = device_id

    def set_topic(self, topic):
        self.topic = topic

    def get_topic(self):
        return self.topic

    def to_dto(self) -> OutputDTO:
        """ convert output object to a data transfer object """
        if self.formatter is None:
            format_dto = "None"
        else:
            format_dto = self.formatter.to_dto()
        return OutputDTO(type=self.name, format=format_dto)



import logging

from .abstractprotocol import AbstractProtocol
from .protocol_helpers import crcPI as crc
from .protocol_helpers import crc8P1 as chk

log = logging.getLogger("pi30revo")

COMMANDS = {
    "PQSE": {
        "name": "PQSE",
        "description": "Read current machine configuration information",
        "help": " -- queries the value of various metrics from the Inverter",
        "type": "QUERY",
        "crctype": "chk",
        "response": [
            [
                "multi",
                [
                    ["option", "Language", ["English", "Chinese"]],
                    [
                        "option",
                        "Work Mode",
                        [
                            "Utility Priority",
                            "Solar Priority",
                            "Battery Priority",
                            "PV+mains power+no external CT",
                            "PV+mains power+external CT",
                        ],
                    ],
                    ["option", "Input Range", ["Wide", "Narrow"]],
                    ["option", "Output Voltage", ["220V", "230V", "240V"]],
                    ["option", "Output Frequency", ["50Hz", "60Hz"]],
                    [
                        "option",
                        "Battery Type",
                        ["Lead Acid", "Lithium", "None", "Custom"],
                    ],
                ],
            ],
            ["float", "Battery Bulk Charge Voltage", "V"],
            ["float", "Battery Float Charge Voltage", "V"],
            ["float", "Battery Low Voltage Alarm Point", "V"],
            ["float", "Battery Low Voltage Cutoff Point", "V"],
            ["int", "Total Charging Current", "A"],
            ["int", "AC Charging Current", "A"],
            ["string", "Year", ""],
            ["string", "Month", ""],
            ["string", "Day", ""],
            ["string", "Hour", ""],
            ["string", "Minute", ""],
            [
                "multi",
                [
                    ["option", "Buzzer", ["Enabled", "Disabled"]],
                    ["option", "Grid", ["Not Connected", "Connected"]],
                    ["option", "Single Phase Parallel", ["Disabled", "Enabled"]],
                    ["option", "Three Phase Parallel", ["Disabled", "Enabled"]],
                    ["string", "Three Phase ID"],
                ],
            ],
            ["float", "Battery Full Recovery Point", "V"],
        ],
        "test_responses": [
            b"(000100 56.0 54.0 44.0 42.0 030 010 2018 06 01 20 00 00 00\x3c\n",
        ],
    },
    "QALL": {
        "name": "QALL",
        "description": "Current Status Data inquiry",
        "help": " -- queries the value of various metrics from the Inverter",
        "type": "QUERY",
        "crctype": "chk",
        "response": [
            ["int", "AC Input Voltage", "V"],
            ["float", "AC Input Frequency", "Hz"],
            ["int", "AC Output Voltage", "V"],
            ["float", "AC Output Frequency", "Hz"],
            ["int", "AC Output Active Power", "W"],
            ["int", "AC Output Load", "%"],
            ["float", "Battery Voltage", "V"],
            ["int", "Battery Capacity", "%"],
            ["int", "Battery Charging Current", "A"],
            ["int", "Battery Discharging Current", "A"],
            ["int", "PV Input Voltage", "V"],
            ["float", "PV Input Current", "A"],
            ["int", "PV Power", "W"],
            ["int", "Daily PV Power Generated", "WH"],
            ["int", "Total PV Power Generated", "kWH"],
            [
                "keyed",
                "Device Mode",
                {
                    "P": "Power on",
                    "S": "Standby",
                    "L": "Line",
                    "B": "Battery",
                    "F": "Fault",
                    "H": "Power saving",
                },
            ],
            [
                "keyed",
                "Warning Code",
                {
                    "00": "No warning",
                    "01": "Low battery",
                    "02": "Mains low voltage",
                    "03": "Mains high voltage",
                    "04": "Overload",
                    "05": "Over temperature",
                    "06": "Fan lock",
                    "07": "Battery overvoltage",
                    "08": "Discharge overcurrent",
                    "21": "PV undervoltage",
                    "22": "PV overvoltage",
                    "23": "PV overcurrent",
                    "24": "PV over temperature",
                    "25": "PV overload",
                    "26": "PV boost failed",
                    "39": "BMS Communication lost",
                },
            ],
            [
                "keyed",
                "Fault Code",
                {
                    "00": "No fault",
                    "01": "BUS overvoltage",
                    "02": "Inverter overvoltage",
                    "03": "Inverter low voltage",
                    "04": "BUS soft start failure",
                    "05": "Overload fault",
                    "06": "Output short circuit",
                    "07": "Low battery voltage failure",
                    "08": "Inverter soft start failure",
                    "09": "BUS low voltage",
                    "10": "Parallel failure",
                    "11": "Over temperature fault",
                    "12": "Battery over-voltage fault",
                    "13": "Phase A lost",
                    "14": "Phase B lost",
                    "15": "Phase C lost",
                    "16": "AC output voltage and frequency setting is different",
                    "17": "AC input voltage and frequency detected different",
                    "18": "Power feedback protection",
                    "19": "Firmware Version inconsistent",
                    "20": "Current sharing fault",
                    "23": "PV overcurrent",
                    "24": "PV over temperature",
                },
            ],
        ],
        "test_responses": [
            b"(000 00.0 000 00.0 0000 000 00.0 000 000 000 000 00.0 0000 000000 000000 S 00 00\x04\n",
        ],
    },
    "QMOD": {
        "name": "QMOD",
        "description": "Mode inquiry",
        "help": " -- queries the Inverter mode",
        "type": "QUERY",
        "crctype": "crc",
        "response": [
            [
                "keyed",
                "Device Mode",
                {
                    "P": "Power on",
                    "S": "Standby",
                    "L": "Line",
                    "B": "Battery",
                    "F": "Fault",
                    "H": "Power saving",
                },
            ]
        ],
        "test_responses": [
            b"(S\x64\x39\r",
        ],
    },
    "QPI": {
        "name": "QPI",
        "description": "Protocol ID inquiry",
        "help": " -- queries the device protocol ID. e.g. PI30 for HS series",
        "type": "QUERY",
        "crctype": "crc",
        "response": [
            ["string", "Protocol ID", ""],
        ],
        "test_responses": [
            b"(PI30\x9a\x0b\r",
        ],
    },
    "QPIGS": {
        "name": "QPIGS",
        "description": "General Status Parameters inquiry",
        "help": " -- queries the value of various metrics from the Inverter",
        "type": "QUERY",
        "crctype": "crc",
        "response": [
            ["float", "AC Input Voltage", "V"],
            ["float", "AC Input Frequency", "Hz"],
            ["float", "AC Output Voltage", "V"],
            ["float", "AC Output Frequency", "Hz"],
            ["int", "AC Output Apparent Power", "VA"],
            ["int", "AC Output Active Power", "W"],
            ["int", "AC Output Load", "%"],
            ["int", "BUS Voltage", "V"],
            ["float", "Battery Voltage", "V"],
            ["int", "Battery Charging Current", "A"],
            ["int", "Battery Capacity", "%"],
            ["int", "Inverter Heat Sink Temperature", "°C"],
            ["float", "PV Power", "W"],
            ["float", "PV Input Voltage", "V"],
            ["float", "Battery Voltage from SCC", "V"],
            ["int", "Daily Power Generated", "WH"],
            [
                "flags",
                "Device Status",
                [
                    "Is SBU Priority Version Added",
                    "Is Configuration Changed",
                    "Is SCC Firmware Updated",
                    "Is Load On",
                    "Is Battery Voltage to Steady While Charging",
                    "Is Charging On",
                    "Is SCC Charging On",
                    "Is AC Charging On",
                ],
            ],
            ["int", "RSV1", "A"],
            ["int", "RSV2", "A"],
            ["int", "PV Input Power", "W"],
            [
                "flags",
                "Device Status2",
                ["Is Charging to Float", "Is Switched On", "Is Reserved"],
            ],
        ],
        "test_responses": [
            b"(221.0 49.9 221.0 49.9 0000 0707 012 401 00.10 012 000 0026 0000 000.0 00.10 00000 00010000 00 00 00000 010\xe9}\r"
        ],
    },
    "QPIRI": {
        "name": "QPIRI",
        "description": "Current Settings inquiry",
        "help": " -- queries the current settings from the Inverter",
        "type": "QUERY",
        "crctype": "crc",
        "response": [
            ["float", "AC Input Voltage", "V"],
            ["float", "AC Input Current", "A"],
            ["float", "AC Output Voltage", "V"],
            ["float", "AC Output Frequency", "Hz"],
            ["float", "AC Output Current", "A"],
            ["int", "AC Output Apparent Power", "VA"],
            ["int", "AC Output Active Power", "W"],
            ["float", "Battery Voltage", "V"],
            ["float", "Battery Recharge Voltage", "V"],
            ["float", "Battery Under Voltage", "V"],
            ["float", "Battery Bulk Charge Voltage", "V"],
            ["float", "Battery Float Charge Voltage", "V"],
            ["option", "Battery Type", ["AGM", "Flooded", "User"]],
            ["int", "Max AC Charging Current", "A"],
            ["int", "Max Charging Current", "A"],
            ["option", "Input Voltage Range", ["Appliance", "UPS"]],
            [
                "option",
                "Output Source Priority",
                ["Utility first", "Solar first", "SBU first"],
            ],
            [
                "option",
                "Charger Source Priority",
                [
                    "Utility first",
                    "Solar first",
                    "Solar + Utility",
                    "Only solar charging permitted",
                ],
            ],
            ["int", "Max Parallel Units", "units"],
            [
                "keyed",
                "Machine Type",
                {"00": "Grid tie", "01": "Off Grid", "10": "Hybrid"},
            ],
            ["option", "Topology", ["transformerless", "transformer"]],
            [
                "option",
                "Output Mode",
                [
                    "single machine output",
                    "parallel output",
                    "Phase 1 of 3 Phase output",
                    "Phase 2 of 3 Phase output",
                    "Phase 3 of 3 Phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output",
                    "unknown output",
                ],
            ],
            ["float", "Battery Redischarge Voltage", "V"],
            [
                "option",
                "PV OK Condition",
                [
                    "As long as one unit of inverters has connect PV, parallel system will consider PV OK",
                    "Only All of inverters have connect PV, parallel system will consider PV OK",
                ],
            ],
            [
                "option",
                "PV Power Balance",
                [
                    "PV input max current will be the max charged current",
                    "PV input max power will be the sum of the max charged power and loads power",
                ],
            ],
        ],
        "test_responses": [
            b"(230.0 21.7 230.0 50.0 21.7 5000 4000 48.0 46.0 42.0 56.4 54.0 0 10 010 1 0 0 6 01 0 0 54.0 0 1\x6F\x7E\r",
        ],
    },
    "QLITH0": {
        "name": "QLITH0",
        "description": "Read lithium battery information",
        "help": " -- queries the value of various metrics from the battery",
        "type": "QUERY",
        "crctype": "chk",
        "response": [
            ["float", "Battery voltage from BMS", "V"],
            ["float", "Battery charging current from BMS", "A"],
            ["float", "Battery discharge current from BMS", "A"],
            ["int", "Battery temperature", "0.1°C"],
            ["int", "Battery capacity from BMS", "%"],
            ["int", "Battery wearout", "%"],
            ["float", "Battery max charging current", "A"],
            ["float", "Battery max discharge current", "A"],
            ["float", "Battery max charge voltage", "V"],
            ["float", "Battery min discharge voltage", "V"],
            [
                "keyed",
                "Fault Code from BMS",
                {
                    "0": "No warning",
                    "1": "Battery overvoltage",
                    "2": "Low battery",
                    "3": "Battery over temperature",
                    "4": "Battery low temperature",
                    "5": "Battery overcurrent",
                    "6": "Battery output short circut",
                },
            ],
            [
                "keyed",
                "Warning Code fom BMS",
                {
                    "0": "No warning",
                    "1": "Battery overvoltage",
                    "2": "Low battery",
                    "3": "Battery over temperature",
                    "4": "Battery low temperature",
                    "5": "Battery overcurrent",
                    "6": "Battery output short circut",
                },
            ],
        ],
        "test_responses": [
            b"(048.8 000.0 000.9 250 024 000 052.5 010.0 053.5 004.4 0 2\xFC\n",
        ],
    },
}


class pi30revo(AbstractProtocol):
    def __str__(self):
        return "PI30 protocol handler for REVO and similar inverters"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI30REVO"
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = ["QPIGS"]
        self.SETTINGS_COMMANDS = ["QPIRI"]
        self.DEFAULT_COMMAND = "QPI"
        # log.info(f'Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands')

    def is_CRC_valid(self, response):
        """
        Test response using CRC approach
        """
        crc_high, crc_low = crc(response[:-3])
        if response[-3:-1] == bytes([crc_high, crc_low]):
            log.debug("CRCs match")
            return True
        return False

    def is_CHK_valid(self, response):
        """
        Test response using CHK approach
        """
        checksum = chk(response[:-2])
        if response[-2:-1] == bytes([checksum]):
            log.debug("Checksums match")
            return True
        return False

    def get_responses(self, response) -> list:
        """
        Split and trim for this protocol is complicated by 2 possible crc approaches
          each with different lengths
        """
        # This protocol responses can either have a CRC or a Checksum..
        # If there is a valid checksum, assume it is using the checksum approach
        if self.is_CHK_valid(response):
            return response[1:-2].split(b" ")
        # Just use the default approach then
        # Trim leading '(' + trailing CRC and \r of response, then split
        if type(response) is str:
            return response[1:-3].split(" ")
        return response[1:-3].split(b" ")

    def check_response_valid(self, response):
        if response is None:
            return False, {"validity check": ["Error: Response was empty", ""]}
        if len(response) <= 3:
            return False, {"validity check": ["Error: Response to short", ""]}

        # This protocol responses can either have a CRC or a Checksum...
        if b"(NAK" in response:
            return False, {"validity check": ["Error: NAK", ""]}

        if self.is_CHK_valid(response) or self.is_CRC_valid(response):
            return True, {}
        else:
            return False, {"validity check": ["Error: Invalid response CRCs", ""]}

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different for PI30REVO
        """
        log.info(
            f"sing protocol {self._protocol_id} with {len(self.COMMANDS)} commands"
        )
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # End of required variables setting

        byte_cmd = bytes(command, "utf-8")
        if (
            self._command_defn
            and self._command_defn.get("crctype") == "chk"
        ):
            log.debug(f"Using CHK checksum approach for command {self._command}")
            checksum = chk(byte_cmd)
            log.debug(f"checksum {checksum}")
            full_command = byte_cmd + bytes([checksum]) + bytes([13])
        else:
            log.debug(f"Using PI30 CRC checksum approach for command {self._command}")
            # calculate the CRC
            crc_high, crc_low = crc(byte_cmd)
            # combine byte_cmd, CRC , return
            full_command = byte_cmd + bytes([crc_high, crc_low, 13])

        log.debug(f"full command: {full_command}")
        return full_command



""" protocol __init__.py """
import importlib
import logging
from enum import auto

from strenum import LowercaseStrEnum

from powermon.errors import ConfigError

log = logging.getLogger("protocols")


class Protocol(LowercaseStrEnum):
    """ enumerate available protocols """
    PI30 = auto()
    PI30MAX = auto()
    VED = auto()


def get_protocol_definition(protocol):
    """
    Get the protocol based on the protocol name
    """

    log.debug("Protocol: %s", protocol)

    protocol_id = protocol.lower()

    match protocol_id:
        case Protocol.PI30:
            from powermon.protocols.pi30 import PI30
            return PI30()
        case Protocol.PI30MAX:
            from powermon.protocols.pi30max import PI30MAX
            return PI30MAX()
        case Protocol.VED:
            from powermon.protocols.ved import VictronEnergyDirect
            return VictronEnergyDirect()
        case _:
            raise ConfigError(f"Invalid protocol_id, no protocol found for: '{protocol_id}'") 
    return None


def list_protocols():
    """ helper function to display a list of supported protocols """
    print("Supported protocols")
    for name in Protocol:
        try:
            _module_class = importlib.import_module("powermon.protocols." + name, ".")
            _module = getattr(_module_class, name)
        except ModuleNotFoundError as exc:
            log.info("Error in module %s: %s", name, exc)
            continue
        except AttributeError as exc:
            log.info("Error in module %s: %s", name, exc)
            continue
        print(f"{name}: {_module()}")



import logging

from .daly import daly

log = logging.getLogger("daly40")

# (AAA BBB CCC DDD EEE
# (000 001 002 003 004


startFlag = bytes.fromhex("A5")


class daly40(daly):
    """
    DALY - Daly40 BMS protocol handler
    """
    def __str__(self):
        return "Daly40 BMS protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"DALY40"
        self.module_address = bytes.fromhex("40")



""" daemon.py """
import logging
from enum import Enum, auto
from time import time

from pydantic import BaseModel
from strenum import LowercaseStrEnum

# Set-up logger
log = logging.getLogger("daemon")


class dummyNotification(Enum):
    READY = auto()
    STATUS = auto()
    STOPPING = auto()
    WATCHDOG = auto()


class DaemonType(LowercaseStrEnum):
    """ Daemon types implemented """
    DISABLED = auto()
    SYSTEMD = auto()


class DaemonDTO(BaseModel):
    """ data transfer ojbect model """
    daemon_type: DaemonType
    enabled: bool
    keepalive: int


class Daemon:
    """ abstraction to support different daemon approaches / solutions """
    def __str__(self):
        if not self.enabled:
            return "Daemon DISABLED"
        return f"Daemon name: {self.type}"

    @classmethod
    def from_config(cls, config=None):
        """ build the object from a config dict """
        log.debug(f"daemon config: {config}")

        if config is None:
            _type = DaemonType.DISABLED
            keepalive = 0
            enabled = False
            log.debug("daemon not configured, disabling")
        if config is not None:
            enabled = True
            _type = config.get("type", None)
            keepalive = config.get("keepalive", 60)
            log.debug(f"got daemon: {_type=}, {keepalive=}")

        return cls(_type=_type, keepalive=keepalive, enabled=enabled)

    def __init__(self, _type, keepalive, enabled):

        self.enabled = enabled
        self.type = _type
        self.keepalive = keepalive
        log.debug(f"got daemon type: {self.type}, keepalive: {self.keepalive}")

        match self.type:
            case DaemonType.SYSTEMD:
                # TODO: this should probably be separate classes / objects
                try:
                    from cysystemd import journal
                    from cysystemd.daemon import Notification, notify

                    self._notify = notify
                    self._journal = journal.write
                    self._Notification = Notification
                except ModuleNotFoundError as exception:
                    print(
                        f"error: {exception}, try 'pip install cysystemd' (which may need 'apt install build-essential libsystemd-dev'), see https://pypi.org/project/cysystemd/ for further info"
                    )
                    exit(1)
            case _:
                self._notify = self._dummyNotify
                self._journal = self._dummyNotify
                self._Notification = dummyNotification
        self.notify(f"got daemon type: {self.type}, keepalive: {self.keepalive}")

    def initialize(self):
        """ Daemon initialization activities """
        if self.enabled:
            # Send READY=1
            self._notify(self._Notification.READY)
            self._lastNotify = time()

    def watchdog(self):
        if self.enabled:
            elapsed = time() - self._lastNotify
            if (elapsed) > self.keepalive:
                self._notify(self._Notification.WATCHDOG)
                self._lastNotify = time()
                self._journal(f"Daemon notify at {self._lastNotify}")

    def notify(self, status="OK"):
        if not self.enabled:
            return
        # Send status
        self._notify(self._Notification.STATUS, status)

    def stop(self):
        if self.enabled:
            # Send stopping
            self._notify(self._Notification.STOPPING)

    def log(self, message=None):
        if self.enabled:
            # Print log message
            if message is not None:
                self._journal(message)

    def _dummyNotify(self, *args, **kwargs):
        # Print log message
        # if args:
        #     print(args[0])
        return

    def to_dto(self):
        """ return the data transfer object version of this object """
        return DaemonDTO(
            daemon_type=self.type,
            enabled=self.enabled,
            keepalive=self.keepalive,
        )



import logging

from mppsolar.protocols.abstractprotocol import AbstractProtocol
from mppsolar.protocols.protocol_helpers import crcPI as crc

log = logging.getLogger("pi30")

SETTER_COMMANDS = {
    "F": {
        "name": "F",
        "description": "Set Device Output Frequency",
        "help": " -- examples: F50 (set output frequency to 50Hz) or F60 (set output frequency to 60Hz)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "F([56]0)$",
    },
    "MCHGC": {
        "name": "MCHGC",
        "description": "Set Max Charging Current (for parallel units)",
        "help": " -- examples: MCHGC040 (set unit 0 to max charging current of 40A), MCHGC160 (set unit 1 to max charging current of 60A)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "MCHGC(\\d\\d\\d)$",
    },
    "MNCHGC": {
        "name": "MNCHGC",
        "description": "Set Utility Max Charging Current (more than 100A) (for 4000/5000)",
        "help": " -- example: MNCHGC1120 (set unit 1 utility max charging current to 120A)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "MNCHGC(\\d\\d\\d\\d)$",
    },
    "MUCHGC": {
        "name": "MUCHGC",
        "description": "Set Utility Max Charging Current",
        "help": " -- example: MUCHGC130 (set unit 1 utility max charging current to 30A)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "MUCHGC(\\d\\d\\d)$",
    },
    "PBCV": {
        "name": "PBCV",
        "description": "Set Battery re-charge voltage",
        "help": " -- example PBCV44.0 - set re-charge voltage to 44V (12V unit: 11V/11.3V/11.5V/11.8V/12V/12.3V/12.5V/12.8V, 24V unit: 22V/22.5V/23V/23.5V/24V/24.5V/25V/25.5V, 48V unit: 44V/45V/46V/47V/48V/49V/50V/51V)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PBCV(\\d\\d\\.\\d)$",
    },
    "PBDV": {
        "name": "PBDV",
        "description": "Set Battery re-discharge voltage",
        "help": " -- example PBDV48.0 - set re-discharge voltage to 48V (12V unit: 00.0V/12V/12.3V/12.5V/12.8V/13V/13.3V/13.5V/13.8V/14V/14.3V/14.5, 24V unit: 00.0V/24V/24.5V/25V/25.5V/26V/26.5V/27V/27.5V/28V/28.5V/29V, 48V unit: 00.0V/48V/49V/50V/51V/52V/53V/54V/55V/56V/57V/58V, 00.0V means battery is full(charging in float mode).)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PBDV(\\d\\d\\.\\d)$",
    },
    "PBFT": {
        "name": "PBFT",
        "description": "Set Battery Float Charging Voltage",
        "help": " -- example PBFT58.0 - set battery float charging voltage to 58V (48.0 - 58.4V for 48V unit)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PBFT(\\d\\d\\.\\d)$",
    },
    "PBT": {
        "name": "PBT",
        "description": "Set Battery Type",
        "help": " -- examples: PBT00 (set battery as AGM), PBT01 (set battery as FLOODED), PBT02 (set battery as USER)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PBT(0[012])$",
    },
    "PCP": {
        "name": "PCP",
        "description": "Set Device Charger Priority",
        "help": " -- examples: PCP00 (set utility first), PCP01 (set solar first), PCP02 (HS only: set solar and utility), PCP03 (set solar only charging)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PCP(0[0123])$",
    },
    "PCVV": {
        "name": "PCVV",
        "description": "Set Battery C.V. (constant voltage) charging voltage",
        "help": " -- example PCVV48.0 - set charging voltage to 48V (48.0 - 58.4V for 48V unit)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PCVV(\\d\\d\\.\\d)$",
    },
    "PE": {
        "name": "PE",
        "description": "Set the enabled state of an Inverter setting",
        "help": " -- examples: PEa - enable a (buzzer) [a=buzzer, b=overload bypass, j=power saving, K=LCD go to default after 1min, u=overload restart, v=overtemp restart, x=backlight, y=alarm on primary source interrupt, z=fault code record]",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PE(.+)$",
    },
    "PD": {
        "name": "PD",
        "description": "Set the disabled state of an Inverter setting",
        "help": " -- examples: PDa - disable a (buzzer) [a=buzzer, b=overload bypass, j=power saving, K=LCD go to default after 1min, u=overload restart, v=overtemp restart, x=backlight, y=alarm on primary source interrupt, z=fault code record]",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PD(.+)$",
    },
    "PF": {
        "name": "PF",
        "description": "Set Control Parameters to Default Values",
        "help": " -- example PF (reset control parameters to defaults)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
    },
    "PGR": {
        "name": "PGR",
        "description": "Set Grid Working Range",
        "help": " -- examples: PCR00 (set device working range to appliance), PCR01 (set device working range to UPS)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PGR(0[01])$",
    },
    "POP": {
        "name": "POP",
        "description": "Set Device Output Source Priority",
        "help": " -- examples: POP00 (set utility first), POP01 (set solar first), POP02 (set SBU priority)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "POP(0[012])$",
    },
    "POPLG": {
        "name": "POPLG",
        "description": "Set Device Operation Logic",
        "help": " -- examples: POPLG00 (set Auto mode), POPLG01 (set Online mode), POPLG02 (set ECO mode)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "POPLG(0[012])$",
    },
    "POPM": {
        "name": "POPM",
        "description": "Set Device Output Mode (for 4000/5000)",
        "help": " -- examples: POPM01 (set unit 0 to 1 - parallel output), POPM10 (set unit 1 to 0 - single machine output), POPM02 (set unit 0 to 2 - phase 1 of 3), POPM13 (set unit 1 to 3 - phase 2 of 3), POPM24 (set unit 2 to 4 - phase 3 of 3)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "POPM(\\d[01234])$",
    },
    "PPCP": {
        "name": "PPCP",
        "description": "Set Parallel Device Charger Priority (for 4000/5000)",
        "help": " -- examples: PPCP000 (set unit 1 to 00 - utility first), PPCP101 (set unit 1 to 01 - solar first), PPCP202 (set unit 2 to 02 - solar and utility), PPCP003 (set unit 0 to 03 - solar only charging)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PPCP(\\d0[0123])$",
    },
    "PPVOKC": {
        "name": "PPVOKC",
        "description": "Set PV OK Condition",
        "help": " -- examples: PPVOKC0 (as long as one unit has connected PV, parallel system will consider PV OK), PPVOKC1 (only if all inverters have connected PV, parallel system will consider PV OK)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PPVOKC([01])$",
    },
    "PSDV": {
        "name": "PSDV",
        "description": "Set Battery Cut-off Voltage",
        "help": " -- example PSDV40.0 - set battery cut-off voltage to 40V (40.0 - 48.0V for 48V unit)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PSDV(\\d\\d\\.\\d)$",
    },
    "PSPB": {
        "name": "PSPB",
        "description": "Set Solar Power Balance",
        "help": " -- examples: PSPB0 (PV input max current will be the max charged current), PSPB1 (PV input max power will be the sum of the max charge power and loads power)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PSPB([01])$",
    },
    "PBATCD": {
        "name": "PBATCD",
        "description": "Battery charge/discharge controlling command",
        "help": " -- examples: PBATCDxxx (please read description, use carefully)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PBATCD([01][01][01])$",
    },
    "DAT": {
        "name": "DAT",
        "description": "Set Date Time",
        "help": " -- examples: DATYYYYMMDDHHMMSS (14 digits after DAT)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "DAT(\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
    "PBATMAXDISC": {
        "name": "PBATMAXDISC",
        "description": "Battery max discharge current",
        "help": " -- examples: PBATMAXDISCxxx (000- disable or 030-150A)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PBATMAXDISC([01]\\d\\d)$",
    },
    "BTA": {
        "name": "BTA",
        "description": "Calibrate inverter battery voltage",
        "help": " -- examples: BTA-01 (reduce inverter reading by 0.05V), BTA+09 (increase inverter reading by 0.45V)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "BTA([-+]0\\d)$",
    },
    "PSAVE": {
        "name": "PSAVE",
        "description": "Save EEPROM changes",
        "help": " -- examples: PSAVE (save changes to eeprom)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
    },
}
QUERY_COMMANDS = {
    "Q1": {
        "name": "Q1",
        "description": "Q1 query",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [1, "Time until the end of absorb charging", "int", "sec"],
            [2, "Time until the end of float charging", "int", "sec"],
            [
                3,
                "SCC Flag",
                "option",
                ["SCC not communicating?", "SCC is powered and communicating"],
            ],
            [4, "AllowSccOnFlag", "bytes.decode", ""],
            [5, "ChargeAverageCurrent", "bytes.decode", ""],
            [6, "SCC PWM temperature", "int", "°C", {"device-class": "temperature"}],
            [7, "Inverter temperature", "int", "°C", {"device-class": "temperature"}],
            [8, "Battery temperature", "int", "°C", {"device-class": "temperature"}],
            [9, "Transformer temperature", "int", "°C", {"device-class": "temperature"}],
            [10, "GPIO13", "int", ""],
            [11, "Fan lock status", "option", ["Not locked", "Locked"]],
            [12, "Not used", "bytes.decode", ""],
            [13, "Fan PWM speed", "int", "%"],
            [14, "SCC charge power", "int", "W", {"icon": "mdi:solar-power", "device-class": "power"}],
            [15, "Parallel Warning", "bytes.decode", ""],
            [16, "Sync frequency", "float", ""],
            [
                17,
                "Inverter charge status",
                "str_keyed",
                {"10": "nocharging", "11": "bulk stage", "12": "absorb", "13": "float"},
                {"icon": "mdi:book-open"},
            ],
        ],
        "test_responses": [
            b"(00000 00000 01 01 00 059 045 053 068 00 00 000 0040 0580 0000 50.00 13\x39\xB9\r",
        ],
    },
    "QBOOT": {
        "name": "QBOOT",
        "description": "DSP Has Bootstrap inquiry",
        "type": "QUERY",
        "response": [["option", "DSP Has Bootstrap", ["No", "Yes"]]],
        "test_responses": [
            "",
        ],
    },
    "QDI": {
        "name": "QDI",
        "description": "Default Settings inquiry",
        "help": " -- queries the default settings from the Inverter",
        "type": "QUERY",
        "response": [
            ["float", "AC Output Voltage", "V"],
            ["float", "AC Output Frequency", "Hz"],
            ["int", "Max AC Charging Current", "A"],
            ["float", "Battery Under Voltage", "V"],
            ["float", "Battery Float Charge Voltage", "V"],
            ["float", "Battery Bulk Charge Voltage", "V"],
            ["float", "Battery Recharge Voltage", "V"],
            ["int", "Max Charging Current", "A"],
            ["option", "Input Voltage Range", ["Appliance", "UPS"]],
            [
                "option",
                "Output Source Priority",
                ["Utility first", "Solar first", "SBU first"],
            ],
            [
                "option",
                "Charger Source Priority",
                [
                    "Utility first",
                    "Solar first",
                    "Solar + Utility",
                    "Only solar charging permitted",
                ],
            ],
            ["option", "Battery Type", ["AGM", "Flooded", "User"]],
            ["option", "Buzzer", ["enabled", "disabled"]],
            ["option", "Power saving", ["disabled", "enabled"]],
            ["option", "Overload restart", ["disabled", "enabled"]],
            ["option", "Over temperature restart", ["disabled", "enabled"]],
            ["option", "LCD Backlight", ["disabled", "enabled"]],
            ["option", "Primary source interrupt alarm", ["disabled", "enabled"]],
            ["option", "Record fault code", ["disabled", "enabled"]],
            ["option", "Overload bypass", ["disabled", "enabled"]],
            ["option", "LCD reset to default", ["disabled", "enabled"]],
            [
                "option",
                "Output mode",
                [
                    "single machine output",
                    "parallel output",
                    "Phase 1 of 3 Phase output",
                    "Phase 2 of 3 Phase output",
                    "Phase 3 of 3 Phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output",
                    "unknown output phase",
                ],
            ],
            ["float", "Battery Redischarge Voltage", "V"],
            [
                "option",
                "PV OK condition",
                [
                    "As long as one unit of inverters has connect PV, parallel system will consider PV OK",
                    "Only All of inverters have connect PV, parallel system will consider PV OK",
                ],
            ],
            [
                "option",
                "PV Power Balance",
                [
                    "PV input max current will be the max charged current",
                    "PV input max power will be the sum of the max charged power and loads power",
                ],
            ],
        ],
        "test_responses": [
            b"(230.0 50.0 0030 42.0 54.0 56.4 46.0 60 0 0 2 0 0 0 0 0 1 1 0 0 1 0 54.0 0 1 000\x9E\x60\r",
        ],
    },
    "QFLAG": {
        "name": "QFLAG",
        "description": "Flag Status inquiry",
        "help": " -- queries the enabled / disabled state of various Inverter settings (e.g. buzzer, overload, interrupt alarm)",
        "type": "QUERY",
        "response": [
            [
                "enflags",
                "Device Status",
                {
                    "a": {"name": "Buzzer", "state": "disabled"},
                    "b": {"name": "Overload Bypass", "state": "disabled"},
                    "j": {"name": "Power Saving", "state": "disabled"},
                    "k": {"name": "LCD Reset to Default", "state": "disabled"},
                    "u": {"name": "Overload Restart", "state": "disabled"},
                    "v": {"name": "Over Temperature Restart", "state": "disabled"},
                    "x": {"name": "LCD Backlight", "state": "disabled"},
                    "y": {
                        "name": "Primary Source Interrupt Alarm",
                        "state": "disabled",
                    },
                    "z": {"name": "Record Fault Code", "state": "disabled"},
                },
            ]
        ],
        "test_responses": [
            b"(EakxyDbjuvz\x2F\x29\r",
        ],
    },
    "QID": {
        "name": "QID",
        "description": "Device Serial Number inquiry",
        "help": " -- queries the device serial number",
        "type": "QUERY",
        "response": [["string", "Serial Number", ""]],
        "test_responses": [
            b"(9293333010501\xBB\x07\r",
        ],
    },
    "QMCHGCR": {
        "name": "QMCHGCR",
        "description": "Max Charging Current Options inquiry",
        "help": " -- queries the maximum charging current setting of the Inverter",
        "type": "QUERY",
        "response": [["string", "Max Charging Current", "A"]],
        "test_responses": [
            b"(010 020 030 040 050 060 070 080 090 100 110 120\x0c\xcb\r",
        ],
    },
    "QMOD": {
        "name": "QMOD",
        "description": "Mode inquiry",
        "help": " -- queries the Inverter mode",
        "type": "QUERY",
        "response": [
            [
                "keyed",
                "Device Mode",
                {
                    "P": "Power on",
                    "S": "Standby",
                    "L": "Line",
                    "B": "Battery",
                    "F": "Fault",
                    "H": "Power saving",
                    "Y": "Bypass",
                },
            ]
        ],
        "test_responses": [
            b"(S\x64\x39\r",
        ],
    },
    "QMN": {
        "name": "QMN",
        "description": "Model Name Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [["bytes.decode", "Model Name", ""]],
        "test_responses": [
            b"(MKS2-8000\xb2\x8d\r",
        ],
    },
    "QGMN": {
        "name": "QGMN",
        "description": "General Model Name Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [["bytes.decode", "General Model Name", ""]],
        "test_responses": [
            b"(044\xc8\xae\r",
        ],
    },
    "QMUCHGCR": {
        "name": "QMUCHGCR",
        "description": "Max Utility Charging Current Options inquiry",
        "help": " -- queries the maximum utility charging current setting of the Inverter",
        "type": "QUERY",
        "response": [["string", "Max Utility Charging Current", "A"]],
        "test_responses": [
            b"(002 010 020 030 040 050 060 070 080 090 100 110 120\xca#\r",
        ],
    },
    "QOPM": {
        "name": "QOPM",
        "description": "Output Mode inquiry",
        "help": " -- queries the output mode of the Inverter (e.g. single, parallel, phase 1 of 3 etc)",
        "type": "QUERY",
        "response": [
            [
                "option",
                "Output mode",
                [
                    "single machine output",
                    "parallel output",
                    "Phase 1 of 3 Phase output",
                    "Phase 2 of 3 Phase output",
                    "Phase 3 of 3 Phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output",
                    "unknown output phase",
                ],
            ]
        ],
        "test_responses": [
            b"",
        ],
    },
    "QPGS": {
        "name": "QPGS",
        "description": "Parallel Information inquiry",
        "help": " -- example: QPGS1 queries the values of various metrics from instance 1 of parallel setup Inverters (numbers from 0)",
        "type": "QUERY",
        "response": [
            ["option", "Parallel instance number", ["Not valid", "valid"]],
            ["string", "Serial number", ""],
            [
                "keyed",
                "Work mode",
                {
                    "P": "Power On Mode",
                    "S": "Standby Mode",
                    "L": "Line Mode",
                    "B": "Battery Mode",
                    "F": "Fault Mode",
                    "H": "Power Saving Mode",
                },
            ],
            [
                "keyed",
                "Fault code",
                {
                    "00": "No fault",
                    "01": "Fan is locked",
                    "02": "Over temperature",
                    "03": "Battery voltage is too high",
                    "04": "Battery voltage is too low",
                    "05": "Output short circuited or Over temperature",
                    "06": "Output voltage is too high",
                    "07": "Over load time out",
                    "08": "Bus voltage is too high",
                    "09": "Bus soft start failed",
                    "11": "Main relay failed",
                    "51": "Over current inverter",
                    "52": "Bus soft start failed",
                    "53": "Inverter soft start failed",
                    "54": "Self-test failed",
                    "55": "Over DC voltage on output of inverter",
                    "56": "Battery connection is open",
                    "57": "Current sensor failed",
                    "58": "Output voltage is too low",
                    "60": "Inverter negative power",
                    "71": "Parallel version different",
                    "72": "Output circuit failed",
                    "80": "CAN communication failed",
                    "81": "Parallel host line lost",
                    "82": "Parallel synchronized signal lost",
                    "83": "Parallel battery voltage detect different",
                    "84": "Parallel Line voltage or frequency detect different",
                    "85": "Parallel Line input current unbalanced",
                    "86": "Parallel output setting different",
                },
            ],
            ["float", "Grid voltage", "V"],
            ["float", "Grid frequency", "Hz"],
            ["float", "AC output voltage", "V"],
            ["float", "AC output frequency", "Hz"],
            ["int", "AC output apparent power", "VA"],
            ["int", "AC output active power", "W"],
            ["int", "Load percentage", "%"],
            ["float", "Battery voltage", "V"],
            ["int", "Battery charging current", "A"],
            ["int", "Battery capacity", "%"],
            ["float", "PV Input Voltage", "V"],
            ["int", "Total charging current", "A"],
            ["int", "Total AC output apparent power", "VA"],
            ["int", "Total output active power", "W"],
            ["int", "Total AC output percentage", "%"],
            [
                "flags",
                "Inverter Status",
                [
                    "Is SCC OK",
                    "Is AC Charging",
                    "Is SCC Charging",
                    "Is Battery Over Voltage",
                    "Is Battery Under Voltage",
                    "Is Line Lost",
                    "Is Load On",
                    "Is Configuration Changed",
                ],
            ],
            [
                "option",
                "Output mode",
                [
                    "single machine",
                    "parallel output",
                    "Phase 1 of 3 phase output",
                    "Phase 2 of 3 phase output",
                    "Phase 3 of 3 phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output",
                    "Unknown Output Mode",
                ],
            ],
            [
                "option",
                "Charger source priority",
                ["Utility first", "Solar first", "Solar + Utility", "Solar only"],
            ],
            ["int", "Max charger current", "A"],
            ["int", "Max charger range", "A"],
            ["int", "Max AC charger current", "A"],
            ["int", "PV input current", "A"],
            ["int", "Battery discharge current", "A"],
            ["float", "Unknown float", ""],
            ["string", "Unknown flags?", ""],
        ],
        "test_responses": [
            b"(1 92931701100510 B 00 000.0 00.00 230.6 50.00 0275 0141 005 51.4 001 100 083.3 002 00574 00312 003 10100110 1 2 060 120 10 04 000\xcc#\r",
            b"(1 92912102100033 B 00 000.0 00.00 120.1 59.99 0048 0000 000 53.1 000 059 000.0 000 00154 00016 000 00000110 7 1 060 120 030 00 000 000.0 00\xe7c\r",
            b"QPGS0?\xda\r",
        ],
        "regex": "QPGS(\\d+)$",
    },
    "QPI": {
        "name": "QPI",
        "description": "Protocol ID inquiry",
        "help": " -- queries the device protocol ID. e.g. PI30 for HS series",
        "type": "QUERY",
        "response": [
            ["string", "Protocol ID", ""],
        ],
        "test_responses": [
            b"(PI30\x9a\x0b\r",
        ],
    },
    "QPIGS": {
        "name": "QPIGS",
        "description": "General Status Parameters inquiry",
        "help": " -- queries the value of various metrics from the Inverter",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [0, "AC Input Voltage", "float", "V", {"icon": "mdi:lightning"}],
            [1, "AC Input Frequency", "float", "Hz"],
            [2, "AC Output Voltage", "float", "V"],
            [3, "AC Output Frequency", "float", "Hz"],
            [4, "AC Output Apparent Power", "int", "VA"],
            [5, "AC Output Active Power", "int", "W"],
            [6, "AC Output Load", "int", "%"],
            [7, "BUS Voltage", "int", "V"],
            [8, "Battery Voltage", "float", "V"],
            [9, "Battery Charging Current", "int", "A"],
            [10, "Battery Capacity", "int", "%"],
            [11, "Inverter Heat Sink Temperature", "int", "°C"],
            [12, "PV Input Current for Battery", "float", "A"],
            [13, "PV Input Voltage", "float", "V"],
            [14, "Battery Voltage from SCC", "float", "V"],
            [15, "Battery Discharge Current", "int", "A"],
            [
                16,
                "Device Status",
                "flags",
                [
                    "Is SBU Priority Version Added",
                    "Is Configuration Changed",
                    "Is SCC Firmware Updated",
                    "Is Load On",
                    "Is Battery Voltage to Steady While Charging",
                    "Is Charging On",
                    "Is SCC Charging On",
                    "Is AC Charging On",
                ],
            ],
            [17, "RSV1", "int", "A"],
            [18, "RSV2", "int", "A"],
            [19, "PV Input Power", "int", "W"],
            [
                20,
                "Device Status2",
                "flags",
                ["Is Charging to Float", "Is Switched On", "Is Reserved"],
            ],
        ],
        "test_responses": [
            b"(000.0 00.0 230.0 49.9 0161 0119 003 460 57.50 012 100 0069 0014 103.8 57.45 00000 00110110 00 00 00856 010\x24\x8c\r",
        ],
    },
    "QPIRI": {
        "name": "QPIRI",
        "description": "Current Settings inquiry",
        "help": " -- queries the current settings from the Inverter",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            ["float", "AC Input Voltage", "V"],
            ["float", "AC Input Current", "A"],
            ["float", "AC Output Voltage", "V"],
            ["float", "AC Output Frequency", "Hz"],
            ["float", "AC Output Current", "A"],
            ["int", "AC Output Apparent Power", "VA"],
            ["int", "AC Output Active Power", "W"],
            ["float", "Battery Voltage", "V"],
            ["float", "Battery Recharge Voltage", "V"],
            ["float", "Battery Under Voltage", "V"],
            ["float", "Battery Bulk Charge Voltage", "V"],
            ["float", "Battery Float Charge Voltage", "V"],
            [
                "option",
                "Battery Type",
                [
                    "AGM",
                    "Flooded",
                    "User",
                    "Pylontech",
                    "Shinheung",
                    "WECO",
                    "Soltaro",
                    "TBD",
                    "LIb-protocol compatible",
                    "3rd party Lithium",
                ],
            ],
            ["int", "Max AC Charging Current", "A"],
            ["int", "Max Charging Current", "A"],
            ["option", "Input Voltage Range", ["Appliance", "UPS"]],
            [
                "option",
                "Output Source Priority",
                ["Utility first", "Solar first", "SBU first"],
            ],
            [
                "option",
                "Charger Source Priority",
                [
                    "Utility first",
                    "Solar first",
                    "Solar + Utility",
                    "Only solar charging permitted",
                ],
            ],
            ["int", "Max Parallel Units", "units"],
            [
                "str_keyed",
                "Machine Type",
                {"00": "Grid tie", "01": "Off Grid", "10": "Hybrid"},
            ],
            ["option", "Topology", ["transformerless", "transformer"]],
            [
                "option",
                "Output Mode",
                [
                    "single machine output",
                    "parallel output",
                    "Phase 1 of 3 Phase output",
                    "Phase 2 of 3 Phase output",
                    "Phase 3 of 3 Phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output",
                    "unknown output",
                ],
            ],
            ["float", "Battery Redischarge Voltage", "V"],
            [
                "option",
                "PV OK Condition",
                [
                    "As long as one unit of inverters has connect PV, parallel system will consider PV OK",
                    "Only All of inverters have connect PV, parallel system will consider PV OK",
                ],
            ],
            [
                "option",
                "PV Power Balance",
                [
                    "PV input max current will be the max charged current",
                    "PV input max power will be the sum of the max charged power and loads power",
                ],
            ],
            ["int", "Max charging time for CV stage", "min"],
            [
                "option",
                "Operation Logic",
                ["Automatic mode", "On-line mode", "ECO mode"],
            ],
        ],
        "test_responses": [
            b"(230.0 21.7 230.0 50.0 21.7 5000 4000 48.0 46.0 42.0 56.4 54.0 0 10 010 1 0 0 6 01 0 0 54.0 0 1\x6F\x7E\r",
            b"(120.0 25.0 120.0 60.0 25.0 3000 3000 48.0 46.0 44.0 58.4 54.4 2 30 060 1 2 0 9 01 0 6 54.0 0 1 000 0\x8f\xed\r",
            b"(230.0 13.0 230.0 50.0 13.0 3000 2400 24.0 23.0 21.0 28.2 27.0 0 30 50 0 2 1 - 01 1 0 26.0 0 0\xb9\xbd\r",
            b"(230.0 21.7 230.0 50.0 21.7 5000 5000 48.0 47.0 46.5 57.6 57.6 5 30 080 0 1 2 1 01 0 0 52.0 0 1\x03$\r",
            b"(230.0 21.7 230.0 50.0 21.7 5000 5000 48.0 47.0 46.5 57.6 57.6 9 30 080 0 1 2 1 01 0 0 52.0 0 1\x9c\x6f\r",
            b"(230.0 34.7 230.0 50.0 34.7 8000 8000 48.0 48.0 42.0 54.0 52.5 2 010 030 1 2 2 9 01 0 0 50.0 0 1 480 0 070\xd9`\r",
        ],
    },
    "QPIWS": {
        "name": "QPIWS",
        "description": "Warning status inquiry",
        "help": " -- queries any active warnings flags from the Inverter",
        "type": "QUERY",
        "response": [
            [
                "stat_flags",
                "Warning",
                [
                    "",
                    "Inverter fault",
                    "Bus over fault",
                    "Bus under fault",
                    "Bus soft fail fault",
                    "Line fail warning",
                    "OPV short warning",
                    "Inverter voltage too low fault",
                    "Inverter voltage too high fault",
                    "Over temperature fault",
                    "Fan locked fault",
                    "Battery voltage too high fault",
                    "Battery low alarm warning",
                    "Reserved",
                    "Battery under shutdown warning",
                    "Reserved",
                    "Overload fault",
                    "EEPROM fault",
                    "Inverter over current fault",
                    "Inverter soft fail fault",
                    "Self test fail fault",
                    "OP DC voltage over fault",
                    "Battery open fault",
                    "Current sensor fail fault",
                    "Battery short fault",
                    "Power limit warning",
                    "PV voltage high warning",
                    "MPPT overload fault",
                    "MPPT overload warning",
                    "Battery too low to charge warning",
                    "",
                    "",
                ],
            ]
        ],
        "test_responses": [
            b"(00000100000000001000000000000000\x56\xA6\r",
        ],
    },
    "QVFW": {
        "name": "QVFW",
        "description": "Main CPU firmware version inquiry",
        "help": " -- queries the main CPU firmware version",
        "type": "QUERY",
        "response": [["string", "Main CPU firmware version", ""]],
        "test_responses": [
            b"(VERFW:00072.70\x53\xA7\r",
        ],
    },
    "QVFW2": {
        "name": "QVFW2",
        "description": "Secondary CPU firmware version inquiry",
        "help": " -- queries the secondary CPU firmware version",
        "type": "QUERY",
        "response": [["string", "Secondary CPU firmware version", ""]],
        "test_responses": [
            b"",
        ],
    },
}


class pi30(AbstractProtocol):
    def __str__(self):
        return "PI30 protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI30"
        self.COMMANDS = QUERY_COMMANDS
        self.COMMANDS.update(SETTER_COMMANDS)
        self.STATUS_COMMANDS = ["QPIGS", "Q1"]
        self.SETTINGS_COMMANDS = ["QPIRI", "QFLAG"]
        self.DEFAULT_COMMAND = "QPI"
        self.PID = "QPI"
        self.ID_COMMANDS = ["QPI", "QGMN", "QMN"]
        # log.info(f'Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands')

    def check_response_valid(self, response):
        if response is None:
            return False, {"validity check": ["Error: Response was empty", ""]}
        if type(response) is dict:
            response["validity check"] = ["Error: incorrect response format", ""]
            return False, response
        if len(response) <= 3:
            return False, {"validity check": ["Error: Response to short", ""]}

        if type(response) is str:
            if "(NAK" in response:
                return False, {"validity check": ["Error: NAK", ""]}
            crc_high, crc_low = crc(response[:-3])
            if [ord(response[-3]), ord(response[-2])] != [crc_high, crc_low]:
                return False, {"validity check": ["Error: Invalid response CRCs", ""]}
        elif type(response) is bytes:
            if b"(NAK" in response:
                return False, {"validity check": ["Error: NAK", ""]}

            crc_high, crc_low = crc(response[:-3])
            if response[-3:-1] != bytes([crc_high, crc_low]):
                return False, {"validity check": ["Error: Invalid response CRCs", ""]}
        log.debug("CRCs match")
        return True, {}



""" protocols / abstractprotocol.py """
import abc
import logging
import re

from mppsolar.protocols.protocol_helpers import crcPI as crc
from powermon.commands.command_definition import CommandDefinition
from powermon.commands.result import ResultType
from powermon.dto.command_definition_dto import CommandDefinitionDTO
from powermon.dto.protocolDTO import ProtocolDTO
from powermon.errors import CommandDefinitionMissing, InvalidResponse, PowermonProtocolError

log = logging.getLogger("AbstractProtocol")


class AbstractProtocol(metaclass=abc.ABCMeta):
    """
    base definition for all protocols
    protocol has:
    - protocol id
    - dict of command definitions
    and functions to:
    - add / remove / count / get command definitions
    - check validity / crc / trim / split response
    """

    def __init__(self) -> None:
        self.command_definitions: dict[str, CommandDefinition] = {}
        self.STATUS_COMMANDS = None
        self.SETTINGS_COMMANDS = None
        self.DEFAULT_COMMAND = None
        self.ID_COMMANDS = None

    @property
    def protocol_id(self) -> bytes:
        """ return the protocol id """
        return self._protocol_id

    @protocol_id.setter
    def protocol_id(self, value):
        self._protocol_id = value

    def add_command_definitions(self, command_definitions_config: dict, result_type: ResultType = None):
        """ Add command definitions from the configuration """
        for command_definition_key in command_definitions_config.keys():
            try:
                log.debug("Attempting to add command_definition_key: %s", command_definition_key)
                _config = command_definitions_config[command_definition_key]
                if result_type is not None:
                    # Adding command definition with supplied type, so override config
                    log.debug("result_type override to %s", result_type)
                    _config["result_type"] = result_type
                command_definition = CommandDefinition.from_config(_config)
                self.command_definitions[command_definition_key] = command_definition
            except ValueError as value_error:
                log.info("couldnt add command definition for code: %s", command_definition_key)
                log.info("error was: %s", value_error)

    def remove_command_definitions(self, commands_to_remove: list = None):
        """ Remove specified command definitions """
        if commands_to_remove is None:
            return
        for command_to_remove in commands_to_remove:
            self.command_definitions.pop(command_to_remove, None)

    def get_command_definition(self, command: str) -> CommandDefinition:
        """ Get the command definition for a given command string """
        # Handle the commands that don't have a regex
        if command in self.command_definitions and self.command_definitions[command].regex is None:
            log.debug("Found command %s in protocol %s", command, self._protocol_id)
            return self.command_definitions[command]

        # Try the regex commands
        for command_code, command_definition in self.command_definitions.items():
            if command_definition.regex is not None:
                log.debug("Regex commands _command: %s", command_code)
                _re = re.compile(command_definition.regex)
                match = _re.match(command)
                if match:
                    log.debug("Matched: %s to: %s value: %s", command, command_definition.code, match.group(1))
                    return command_definition
        log.info("No command_defn found for %s", command)
        raise CommandDefinitionMissing(f"No command definition found for command: {command}")

    def check_definitions_count(self, expected=None):
        """ check and report number of command definitions, error if 0 """
        definitions_count = len(self.command_definitions)
        if definitions_count == 0:
            raise PowermonProtocolError(f"Attempted to load protocol '{self.protocol_id}' which has no valid commands")
        if expected is None:
            log.info("Using protocol:%s with %i commands (%s)", self.protocol_id, definitions_count, ', '.join(self.command_definitions.keys()))
            return
        if expected == definitions_count:
            log.info("Using protocol:%s found %i commands as expected (%s)", self.protocol_id, definitions_count, ', '.join(self.command_definitions.keys()))
            return
        else:
            raise PowermonProtocolError(f"Loaded protocol '{self.protocol_id}' but found {definitions_count} commands, expected {expected}")

    def get_command_definition_dtos(self) -> dict[str, CommandDefinitionDTO]:
        """ convert all associated command objects to data transfer objects """
        command_dtos: dict[str, CommandDefinitionDTO] = {}
        for command_tuple in self.command_definitions.items():
            command_dtos[command_tuple[0]] = command_tuple[1].to_dto()
        return command_dtos

    def list_commands(self) -> dict[str, CommandDefinition]:
        """ list available commands for the protocol """
        if self.protocol_id is None:
            log.error("Attempted to list commands with no protocol defined")
            raise ValueError("Attempted to list commands with no protocol defined")
        return self.command_definitions

    def get_full_command(self, command: str) -> bytes:
        """ generate the full command including crc and \n as needed """
        log.info("Using protocol: %s with %i commands", self.protocol_id, len(self.command_definitions))
        byte_cmd = bytes(command, "utf-8")
        # calculate the CRC
        crc_high, crc_low = crc(byte_cmd)
        # combine byte_cmd, CRC , return
        full_command = byte_cmd + bytes([crc_high, crc_low, 13])
        log.debug("full command: %s", full_command)
        return full_command

    def check_valid(self, response: str, command_definition: CommandDefinition = None) -> bool:
        """ check response is valid """
        log.debug("check valid for %s, definition: %s", response, command_definition)
        if response is None:
            raise InvalidResponse("Response is None")
        if len(response) <= 3:
            raise InvalidResponse("Response is too short")
        return True

    def check_crc(self, response: str, command_definition: CommandDefinition = None) -> bool:
        """ crc check, needs override in protocol """
        log.debug("no check crc for %s, definition: %s", response, command_definition)
        return True

    def trim_response(self, response: str, command_definition: CommandDefinition = None) -> str:
        """ Remove extra characters from response """
        log.debug("trim %s, definition: %s", response, command_definition)
        return response[1:-3]

    def split_response(self, response: str, command_definition: CommandDefinition = None) -> list | dict:
        """ split response into individual items, return as ordered list or keyed dict """
        result_type = getattr(command_definition, "result_type", None)
        log.debug("splitting %s, result_type %s", response, result_type)
        match result_type:
            case None:
                responses = []
            case ResultType.ACK | ResultType.SINGLE | ResultType.MULTIVALUED:
                responses = response
            case ResultType.SLICED:
                responses = []
                for position in range(command_definition.reading_definition_count()):
                    rd = command_definition.get_reading_definition(position=position)
                    responses.append(response[rd.slice_array[0]:rd.slice_array[1]])
            case ResultType.VED_INDEXED:
                # build a list of (index,value) tuples
                responses = []
                for item in response.split(b'\r\n'):
                    try:
                        key, value = item.split(b'\t')
                    except ValueError:
                        continue
                    if isinstance(key, bytes):
                        key = key.decode()
                    responses.append((key.strip(), value.strip()))
            case _:
                responses = response.split()
        log.debug("responses: '%s'", responses)
        return responses

    def to_dto(self) -> ProtocolDTO:
        """ convert protocol object to data transfer object """
        dto = ProtocolDTO(protocol_id=self.protocol_id, commands=self.get_command_definition_dtos())
        return dto



import logging
from dataclasses import dataclass
from typing import Optional

import paho.mqtt.client as mqtt_client

# Set-up logger
log = logging.getLogger("mqttbroker")


@dataclass
class MqttBroker:
    name: str
    port: int = 1883
    username: Optional[str] = None
    password: Optional[str] = None
    results_topic: Optional[str] = None
    mqttc: mqtt_client.Client = mqtt_client.Client()
    _isConnected: bool = False

    def on_connect(self, client, userdata, flags, rc):
        # 0: Connection successful
        # 1: Connection refused - incorrect protocol version
        # 2: Connection refused - invalid client identifier
        # 3: Connection refused - server unavailable
        # 4: Connection refused - bad username or password
        # 5: Connection refused - not authorised
        # 6-255: Currently unused.
        connection_result = [
            "Connection successful",
            "Connection refused - incorrect protocol version",
            "Connection refused - invalid client identifier",
            "Connection refused - server unavailable",
            "Connection refused - bad username or password",
            "Connection refused - not authorised",
        ]
        log.debug(
            f"MqttBroker connection returned result: {rc} {connection_result[rc]}"
        )
        if rc == 0:
            self._isConnected = True
            return
        self._isConnected = False

    def on_disconnect(self, client, userdata, rc):
        log.error(f"Disconnection returned result: {rc}")
        self._isConnected = False

    def connect(self):

        self.mqttc.on_connect = self.on_connect
        self.mqttc.on_disconnect = self.on_disconnect
        # if a name is supplied, connect
        if self.name:
            self.mqttc.connect(self.name, self.port, keepalive=60)

    def start(self):
        if self._isConnected:
            self.mqttc.loop_start()

    def stop(self):
        if self.name:
            self.mqttc.loop_stop()
            if self._isConnected:
                self.mqttc.disconnect

    def set(self, variable, value):
        setattr(self, variable, value)

    def update(self, variable, value):
        # only override if value is not None
        if value is None:
            return
        setattr(self, variable, value)

    def subscribe(self, topic, callback):
        # subscribe to mqtt topic
        if not self.name:
            return
        # check if connected, connect if not
        if not self._isConnected:
            log.debug("Not connected, connecting")
            self.connect
        # Register callback
        self.mqttc.on_message = callback
        if self._isConnected:
            # Subscribe to command topic
            self.mqttc.subscribe(topic, qos=0)


if __name__ == "__main__":
    broker = MqttBroker("brokername")
    print(broker)
    broker.name = "test1"
    print(broker)



""" mppsolar / protocols / pi17infini.py """
import logging

from .pi17 import pi17


log = logging.getLogger("pi17infini")

QUERY_COMMANDS = {}
SETTER_COMMANDS = {
    "ACCT": {
        "name": "ACCT",
        "description": "Set AC charge time range",
        "help": " -- examples: ACCT2200,2259,2300,2359 (Sets primary time range from 22:00 to 22:59 and secondary time range from 23:00 to 23:59. End minute is inclusive)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "ACCT((2[0-3]|[01]?[0-9])([0-5]?[0-9]),(2[0-3]|[01]?[0-9])([0-5]?[0-9]),(2[0-3]|[01]?[0-9])([0-5]?[0-9]),(2[0-3]|[01]?[0-9])([0-5]?[0-9]))$",
    },
}
COMMANDS_TO_REMOVE = []


class pi17infini(pi17):
    """ pi17 protocol for Infini-Solar inverters (2022 specs) """
    def __str__(self):
        return "PI17INFINI protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI17INFINI"
        # Add pi17infini specific commands to pi17 commands
        self.COMMANDS.update(QUERY_COMMANDS)
        # Add pi17infini specific setter commands
        self.COMMANDS.update(SETTER_COMMANDS)
        # remove and unwanted pi30 commands
        for item in COMMANDS_TO_REMOVE:
            self.COMMANDS.pop(item, None)



""" protocols / ved.py """
import logging
from enum import Enum
from struct import unpack

from mppsolar.protocols.protocol_helpers import vedHexChecksum
from powermon.commands.command_definition import CommandDefinition
from powermon.commands.reading_definition import ReadingType, ResponseType
from powermon.commands.result import ResultType
from powermon.errors import CommandError, InvalidCRC, InvalidResponse
from powermon.protocols.abstractprotocol import AbstractProtocol

log = logging.getLogger("ved")


class VictronCommandType(Enum):
    """ enum of valid types of Results """
    PING = 1
    GET_FW = 3
    GET_ID = 4
    RESTART = 6
    GET = 7
    SET = 8
    ASYNC = 'A'
    LISTEN = 'L'


COMMANDS = {
    "vedtext": {
        "name": "vedtext",
        "description": "VE Direct Text",
        "help": " -- the output of the VE Direct text protocol",
        "device_command_type": VictronCommandType.LISTEN,
        "result_type": ResultType.VED_INDEXED,
        "reading_definitions": [
            {"index": "V", "description": "Main or channel 1 battery voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "V2", "description": "Channel 2 battery voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "V3", "description": "Channel 3 battery voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "VS", "description": "Auxiliary starter voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "VM", "description": "Mid-point voltage of the battery bank", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "DM", "description": "Mid-point deviation of the battery bank", "reading_type": ReadingType.PERCENTAGE, "response_type": ResponseType.FLOAT},
            {"index": "VPV", "description": "Panel voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "PPV", "description": "Panel power", "reading_type": ReadingType.WATTS, "response_type": ResponseType.FLOAT},
            {"index": "I", "description": "Main or channel 1 battery current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "I2", "description": "Channel 2 battery current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "I3", "description": "Channel 3 battery current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "IL", "description": "Load current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "LOAD", "description": "Load output state ON/OFF", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.BYTES},
            {"index": "T", "description": "Battery temperature", "reading_type": ReadingType.TEMPERATURE, "response_type": ResponseType.FLOAT},
            {"index": "P", "description": "Instantaneous power", "reading_type": ReadingType.WATTS, "response_type": ResponseType.FLOAT},
            {"index": "CE", "description": "Consumed Amp Hours", "reading_type": ReadingType.ENERGY, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "SOC", "description": "State-of-charge", "reading_type": ReadingType.PERCENTAGE, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"index": "TTG", "description": "Time-to-go", "reading_type": ReadingType.TIME_MINUTES, "response_type": ResponseType.FLOAT},
            {"index": "Alarm", "description": "Alarm condition active", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "Relay", "description": "Relay state", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "AR", "description": "Alarm reason", "reading_type": ReadingType.MESSAGE,
             "response_type": ResponseType.BIT_ENCODED,
             "options": {0: "No alarm",
                         1: "Low Voltage",
                         2: "High Voltage",
                         4: "Low SOC",
                         8: "Low Starter Voltage",
                         16: "High Starter Voltage",
                         32: "Low Temperature",
                         64: "High Temperature",
                         128: "Mid Voltage",
                         256: "Overload",
                         512: "DC-ripple",
                         1024: "Low V AC out",
                         2048: "High V AC out",
                         4096: "Short Circuit",
                         8192: "BMS Lockout"}},
            {"index": "OR", "description": "Off reason", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "H1", "description": "Depth of the deepest discharge", "reading_type": ReadingType.ENERGY, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "H2", "description": "Depth of the last discharge", "reading_type": ReadingType.ENERGY, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "H3", "description": "Depth of the average discharge", "reading_type": ReadingType.ENERGY, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "H4", "description": "Number of charge cycles", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.INT},
            {"index": "H5", "description": "Number of full discharges", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.INT},
            {"index": "H6", "description": "Cumulative Amp Hours drawn", "reading_type": ReadingType.ENERGY, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "H7", "description": "Minimum main battery voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "H8", "description": "Maximum main battery voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "H9", "description": "Number of seconds since last full charge", "reading_type": ReadingType.TIME_SECONDS, "response_type": ResponseType.FLOAT},
            {"index": "H10", "description": "Number of automatic synchronizations", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.INT},
            {"index": "H11", "description": "Number of low main voltage alarms", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.INT},
            {"index": "H12", "description": "Number of high main voltage alarms", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.INT},
            {"index": "H13", "description": "Number of low auxiliary voltage alarms", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.INT},
            {"index": "H14", "description": "Number of high auxiliary voltage alarms", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.INT},
            {"index": "H15", "description": "Minimum auxiliary battery voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "H16", "description": "Maximum auxiliary battery voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/1000"},
            {"index": "H17", "description": "Amount of discharged energy", "reading_type": ReadingType.KILOWATT_HOURS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "H18", "description": "Amount of charged energy", "reading_type": ReadingType.KILOWATT_HOURS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "H19", "description": "Yield total - user resettable counter", "reading_type": ReadingType.KILOWATT_HOURS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "H20", "description": "Yield today", "reading_type": ReadingType.KILOWATT_HOURS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "H21", "description": "Maximum power today", "reading_type": ReadingType.WATTS, "response_type": ResponseType.FLOAT},
            {"index": "H22", "description": "Yield yesterday", "reading_type": ReadingType.KILOWATT_HOURS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "H23", "description": "Maximum power yesterday", "reading_type": ReadingType.WATTS, "response_type": ResponseType.FLOAT},
            {"index": "ERR", "description": "Error code", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "CS", "description": "State of operation", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "BMV", "description": "Model description", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "FW", "description": "Firmware version 16 bit", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "FWE", "description": "Firmware version 24 bit", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "PID", "description": "Product ID", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "SER#", "description": "Serial number", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "HSDS", "description": "Day sequence number 0..364", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "MODE", "description": "Device mode", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "AC_OUT_V", "description": "AC output voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "AC_OUT_I", "description": "AC output current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"index": "AC_OUT_S", "description": "AC output apparent power", "reading_type": ReadingType.APPARENT_POWER, "response_type": ResponseType.FLOAT},
            {"index": "WARN", "description": "Warning reason", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "MPPT", "description": "Tracker operation mode", "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.STRING},
            {"index": "Checksum", "description": "Checksum", "reading_type": ReadingType.HEX_STR, "response_type": ResponseType.HEX_CHAR},
            {"index": "DC_IN_V", "description": "DC input voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "DC_IN_I", "description": "DC input current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/10"},
            {"index": "DC_IN_S", "description": "DC input power", "reading_type": ReadingType.WATTS, "response_type": ResponseType.FLOAT},
            {"index": "MON", "description": "DC Monitor Mode", "reading_type": ReadingType.MESSAGE,
             "response_type": ResponseType.OPTION,
             "options": {"-9": "Solar charger",
                         "-8": "Wind turbine",
                         "-7": "Shaft generator",
                         "-6": "Alternator",
                         "-5": "Fuel cell",
                         "-4": "Water generator",
                         "-3": "DC/DC charger",
                         "-2": "AC charger",
                         "-1": "Generic source",
                         "0": "Battery monitor (BMV)",
                         "1": "Generic load",
                         "2": "Electric drive",
                         "3": "Fridge",
                         "4": "Water pump",
                         "5": "Bilge pump",
                         "6": "DC system",
                         "7": "Inverter",
                         "8": "Water heater"}}
        ],
        "test_responses": [
            b"H1\t-32914\r\nH2\t0\r\nH3\t0\r\nH4\t0\r\nH5\t0\r\nH6\t-35652\r\nH7\t12041\r\nH8\t14282\r\nH9\t0\r\nH10\t0\r\nH11\t0\r\nH12\t0\r\nH15\t-22\r\nH16\t0\r\nH17\t46\r\nH18\t48\r\nChecksum\t\x1a\r\nPID\t0xA389\r\nV\t12865\r\nVS\t-14\r\nI\t0\r\nP\t0\r\nCE\t0\r\nSOC\t1000\r\nTTG\t-1\r\nAlarm\tOFF\r\nAR\t140\r\nBMV\tSmartShunt 500A/50mV\r\nFW\t0405\r\nChecksum\tL\r\n",
            b"\x00L\r\nH1\t-32914\r\nH2\t0\r\nH3\t0\r\nH4\t0\r\nH5\t0\r\nH6\t-35652\r\nH7\t12041\r\nH8\t14282\r\nH9\t0\r\nH10\t0\r\nH11\t0\r\nH12\t0\r\nH15\t-22\r\nH16\t0\r\nH17\t46\r\nH18\t48\r\nChecksum\t\x1a\r\nPID\t0xA389\r\nV\t12868\r\nVS\t-12\r\nI\t0\r\nP\t0\r\nCE\t0\r\nSOC\t1000\r\nTTG\t-1\r\nAlarm\tOFF\r\nAR\t0\r\nBMV\tSmartShunt 500A/50mV\r\nFW\t0405\r\nChecksum\tK\r",
            b"\nH1\t-32914\r\nH2\t0\r\nH3\t0\r\nMON\t0\r\nH4\t0\r\nH5\t0\r\nH6\t-35652\r\nH7\t12041\r\nH8\t14282\r\nH9\t0\r\nH10\t0\r\nH11\t0\r\nH12\t0\r\nH15\t-22\r\nH16\t0\r\nH17\t46\r\nH18\t48\r\nChecksum\t\x1a\r\nPID\t0xA389\r\nV\t12868\r\nVS\t-13\r\nI\t0\r\nP\t0\r\nCE\t0\r\nSOC\t1000\r\nTTG\t-1\r\nAlarm\tOFF\r\nAR\t0\r\nBMV\tSmartShunt 500A/50mV\r\nFW\t0405\r\nChecksum\tJ\r",
        ],
    },
    "batteryCapacity": {
        "name": "batteryCapacity",
        "description": "Battery Capacity",
        "help": " -- display the Battery Capacity",
        "device_command_type": VictronCommandType.GET,
        "device_command_code": "1000",  # or should be the more accurate 1000
        "result_type": ResultType.SLICED,
        "reading_definitions": [
            {"description": "Command type", "slice": [0, 1], "reading_type": ReadingType.MESSAGE, "response_type": ResponseType.INT},
            {"description": "Command", "slice": [1, 5], "reading_type": ReadingType.HEX_STR, "response_type": ResponseType.LE_2B_S},
            {"description": "Command response flag", "slice": [5, 7], "reading_type": ReadingType.MESSAGE,
                "response_type": ResponseType.OPTION,
                "options": {"00": "OK",
                            "01": "Unknown ID",
                            "02": "Not supported",
                            "04": "Parameter Error"}},
            {"description": "Battery Capacity", "slice": [7, 11], "reading_type": ReadingType.ENERGY, "response_type": ResponseType.LE_2B_S},
        ],
        "test_responses": [
            b":70010007800C6\n",
            b"\x00\x1a:70010007800C6\n",
            b"70010007800C6\n",  # this one will error
        ],
    },
}


class VictronEnergyDirect(AbstractProtocol):
    """
    VictronEnergyDirect - VEDirect protocol handler
    """

    def __str__(self):
        return "VED protocol handler for Victron direct SmartShunts"

    def __init__(self) -> None:
        super().__init__()
        self.protocol_id = b"VED"
        self.add_command_definitions(COMMANDS)
        self.STATUS_COMMANDS = [
            "vedtext",
        ]
        self.SETTINGS_COMMANDS = [
            "",
        ]
        self.DEFAULT_COMMAND = "vedtext"
        self.check_definitions_count(expected=2)

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different for VEDirect
        """
        log.info("Using protocol %s with %i commands", self.protocol_id, len(self.command_definitions))

        command_definition : CommandDefinition = self.get_command_definition(command)
        if command_definition is None:
            return None

        # VEDHEX
        # : start of command
        # 7 Get
        # 0000 id of the value to get
        # 00 flags
        # 00 cs
        # \n
        # eg b':70010003E\n' = get battery capacity id = 0x1000 = 0010 little endian

        command_type = command_definition.device_command_type
        match command_type:
            case VictronCommandType.GET:
                # command components
                raw_command_code = command_definition.device_command_code  # eg 1000 for batteryCapacity
                if raw_command_code is None:
                    raise CommandError(f"device_command_code not found for {command=} - check protocol definition for this command")
                command_code = f"{unpack('<h', bytes.fromhex(raw_command_code))[0]:04X}"
                flags = "00"

                # build command
                cmd = f"{command_type.value}{command_code}{flags}"
                # pad cmd and convert to bytes and determine checksum
                checksum = vedHexChecksum(bytes.fromhex(f"0{cmd}"))

                # build full command
                cmd = f":{cmd}{checksum:02X}\n".encode()
                log.debug("full command: %s", cmd)
                return cmd
            case VictronCommandType.LISTEN:
                # Just listen - dont need to send a command
                log.debug("command is LISTEN type so returning %s", command_type)
                return command_type
        raise CommandError(f"unable to generate full command for {command}, type {command_type} - is the definition wrong or CommandType not implemented?")

    def check_valid(self, response: str, command_definition: CommandDefinition = None) -> bool:
        """ check response is valid """
        log.debug("check valid for %s, definition: %s", response, command_definition)
        if response is None:
            raise InvalidResponse("Response is None")
        if len(response) <= 3:
            raise InvalidResponse("Response is too short")
        command_type = command_definition.device_command_type
        match command_type:
            case VictronCommandType.GET:
                if response.count(b':') != 1:
                    raise InvalidResponse("Response incomplete - missing ':'")
        return True

    def check_crc(self, response: str, command_definition: CommandDefinition = None) -> bool:
        """ crc check, needs override in protocol """
        log.debug("checking crc for %s", response)
        command_type = command_definition.device_command_type
        match command_type:
            case VictronCommandType.GET:
                # HEX protocol response
                log.debug("checking validity of '%s'", response)
                _r = response.split(b":")[1][:-1].decode()
                # print(f"trimmed response {_r}")
                _r = f"0{_r}"
                # print(f"padded response {_r}")
                _r = bytes.fromhex(_r)
                # print(f"bytes response {_r}")
                data = _r[:-1]
                checksum = _r[-1:][0]
                expected_checksum = vedHexChecksum(data)
                if expected_checksum == checksum:
                    log.debug("VED Hex Checksum matches in response '%s' checksum:'%s'", response, checksum)
                    return True
                else:
                    # print("VED Hex Checksum does not match")
                    raise InvalidCRC(f"response has invalid CRC - got '\\x{checksum:02x}', calculated '\\x{expected_checksum:02x}")
            case VictronCommandType.LISTEN:
                return True
        return True

    def trim_response(self, response: str, command_definition: CommandDefinition = None) -> str:
        """ Remove extra characters from response """
        log.debug("response: %s", response)
        command_type = command_definition.device_command_type
        _ret = None
        match command_type:
            case VictronCommandType.GET:
                # HEX response, e.g. b":70010007800C6\n"
                _ret = response.split(b":")[1][:-3]
            case VictronCommandType.LISTEN:
                # VEDTEXT response, return the lot
                _ret = response
        log.debug("trim_response: %s", _ret)
        return _ret



""" commands / command_definition.py """
import logging
import re

from powermon.commands.reading_definition import ReadingDefinition, ReadingType
from powermon.commands.result import ResponseType, ResultType
from powermon.dto.command_definition_dto import CommandDefinitionDTO

log = logging.getLogger("CommandDefinition")


class CommandDefinition:
    """ object the contains the definition and other metadata about a command, including: 
    - code
    - regex (opt)
    - description
    - result type
    - reading definitions
    - test responses
    """
    def __str__(self):
        return f"CommandDefinition: {self.code=}, {self.description=}, {self.result_type=}"

    def __init__(self, code, description, help_text: str, result_type : ResultType, reading_definitions, test_responses: list = None, regex: str = None):
        """ init CommandDefinition class """
        if reading_definitions is None or len(reading_definitions) == 0:
            raise ValueError(f"reading definitions cannot be None for command_code: {code}")

        self.code = code
        self.description = description
        self.help_text = help_text
        self.result_type : ResultType = result_type
        self.reading_definitions : dict[int, ReadingDefinition] = reading_definitions  # TODO: this is incorrect, needs positional and str indexes as well
        self.test_responses : list[bytes] = test_responses
        self.regex : str | None = regex
        self.device_command_type = None
        self.device_command_code : str = None

    @classmethod
    def from_config(cls, protocol_dictionary : dict) -> "CommandDefinition":
        """ build command definition object from config dict """
        code = protocol_dictionary.get("name")
        description = protocol_dictionary.get("description")
        help_text = protocol_dictionary.get("help_text")
        test_responses = protocol_dictionary.get("test_responses")
        regex = protocol_dictionary.get("regex", None)
        result_type = protocol_dictionary.get("result_type")
        match result_type:
            case ResultType.ACK:
                # All ResultType.ACK are the same, so put config here instead of duplicating it in the protocol
                log.debug("ResultType.ACK so defaulting reading_definitions")
                reading_definitions : dict[int, ReadingDefinition] = \
                    ReadingDefinition.multiple_from_config([{"description": description, "response_type": ResponseType.ACK, "reading_type": ReadingType.ACK}])
                test_responses = [b"(NAK\x73\x73\r", b"(ACK\x39\x20\r",]
            case _:
                reading_definitions : dict[int, ReadingDefinition] = \
                    ReadingDefinition.multiple_from_config(protocol_dictionary.get("reading_definitions"))

        log.debug("code: %s description: %s reading_definitions: %s", code, description, reading_definitions)
        _command_definition = cls(
            code=code, description=description, help_text=help_text, result_type=result_type,
            reading_definitions=reading_definitions, test_responses=test_responses,
            regex=regex
        )
        _command_definition.device_command_type = protocol_dictionary.get("device_command_type")
        _command_definition.device_command_code = protocol_dictionary.get("device_command_code")
        return _command_definition

    def to_dto(self) -> CommandDefinitionDTO:
        """ convert command definition object to data transfer object """
        return CommandDefinitionDTO(
            command_code=self.code,
            description=self.description,
            help_text=self.help_text,
            result_type=str(self.result_type),
            # responses=self.response_definitions, #TODO: make DTOs for the response definitions
            # test_responses=self.test_responses,
            regex=self.regex
        )

    def is_command_code_valid(self, command_code : str) -> bool:
        """ determines if a command code is valid """
        if self.regex is None:
            return self.code == command_code
        return re.match(self.regex, command_code) is not None

    def get_reading_definition(self, lookup=None, position=0) -> ReadingDefinition:
        """ return the reading definition that corresponds to lookup """
        log.debug("looking for reading definition with: %s, result_type is: %s", lookup, self.result_type)
        if self.reading_definitions is None:
            result = None
        match self.result_type:
            case ResultType.ACK | ResultType.SINGLE | ResultType.MULTIVALUED:
                result = self.reading_definitions[0]
            case ResultType.ORDERED | ResultType.SLICED:
                result = self.reading_definitions[position]
            case ResultType.VED_INDEXED:
                try:
                    result = self.reading_definitions[lookup]
                except KeyError:
                    log.warning("no reading definition found for key: %s", lookup)
                    result = None
            case _:
                print(f"no get_reading_definition for {self.result_type=}")
                exit()
        log.debug("found reading definition: %s", result)
        return result

    def reading_definition_count(self) -> int:
        """ return the number of reading_definitions for this command_definition """
        return 0 if self.reading_definitions is None else len(self.reading_definitions)



""" formats / table.py """
import logging

from powermon.commands.reading import Reading
from powermon.commands.result import Result
from powermon.formats.abstractformat import (AbstractFormat,
                                             get_max_response_lengths, pad)

log = logging.getLogger("Table")


class Table(AbstractFormat):
    """ table formatter - formats results in a table suitable for std out """

    def __str__(self):
        return "outputs the results to standard out in a table (optionally formatted with line art boxes)"

    def __init__(self, formatConfig):
        super().__init__(formatConfig)
        self.name = "table"
        self.extra_info = formatConfig.get("extra_info", False)
        self.draw_lines = formatConfig.get("draw_lines", False)
        # self.command_description = "unknown command"

    # def set_command_description(self, command_description):
    #     self.command_description = command_description

    def format(self, command, result: Result, device_info) -> list[str]:
        log.info("Using output formatter: %s", self.name)

        _result = []
        filtered_responses: list[Reading] = []

        # check for error in result
        if result.error:
            filtered_responses.append(Reading(data_name="Error Count", data_value=len(result.error_messages), data_unit=""))
            for i, message in enumerate(result.error_messages):
                filtered_responses.append(Reading(data_name=f"Error #{i}", data_value=f"{message}", data_unit=""))

        if len(result.readings) == 0:
            filtered_responses.append(Reading(data_name="Error", data_value="No readings in result", data_unit=""))

        filtered_responses.extend(self.format_and_filter_data(result))
        log.debug("displayData: %s", "\n".join((str(a) for a in filtered_responses)))

        # build header
        # command_code = result.command_definition.code

        # Determine column widths
        _pad = 1

        width_p, width_v, width_u = get_max_response_lengths(filtered_responses)
        # Width of parameter column
        width_p += _pad
        if width_p < 9 + _pad:
            width_p = 9 + _pad
        # Width of value column
        width_v += _pad
        if width_v < 6 + _pad:
            width_v = 6 + _pad
        # Width of units column
        width_u += _pad
        if width_u < 5 + _pad:
            width_u = 5 + _pad
        # Total line length
        line_length = width_p + width_v + width_u + 7
        # Check if command / description line is longer and extend line if needed
        cmd_str = f"Command: {command.code} - {command.command_definition.description}"
        width_c = len(cmd_str)
        # log.debug(f"{width_c=}, {line_length=}, {width_p=}, {width_v=}, {width_u=}")
        if line_length < (width_c + 7):
            line_length = width_c + 7
        # Check if columns too short and expand units if needed
        if (width_p + width_v + width_u + 7) <= line_length:
            width_u = line_length - (width_p + width_v + 7)
        # log.debug(f"{width_c=}, {line_length=}, {width_p=}, {width_v=}, {width_u=}")

        # print header
        if self.draw_lines:
            _result.append("\u2554" + ("\u2550" * (line_length - 2)) + "\u2557")
            _result.append(f"\u2551 {cmd_str}" + (" " * (line_length - len(cmd_str) - 3)) + "\u2551")
        else:
            _result.append("-" * (line_length))
            _result.append(f"{cmd_str}" + (" " * (line_length - len(cmd_str) - 2)))
            _result.append("-" * (line_length))

        # print separator
        if self.draw_lines:
            _result.append("\u2560" + ("\u2550" * (width_p + 1)) + "\u2564" + ("\u2550" * (width_v + 1)) + "\u2564" + ("\u2550" * (width_u + 1)) + "\u2563")
        # print column headings
        if self.draw_lines:
            _result.append(f"\u2551 {pad('Parameter', width_p)}\u2502 {pad('Value', width_v)}\u2502 {pad('Unit', width_u)}\u2551")
        else:
            _result.append(f"{pad('Parameter', width_p)} {pad('Value', width_v)} {pad('Unit', width_u)}")
        # print separator
        if self.draw_lines:
            _result.append("\u255f" + ("\u2500" * (width_p + 1)) + "\u253c" + ("\u2500" * (width_v + 1)) + "\u253c" + ("\u2500" * (width_u + 1)) + "\u2562")

        # print data
        for response in filtered_responses:
            name = response.get_data_name()
            value = response.get_data_value()
            unit = response.get_data_unit()
            if self.draw_lines:
                _result.append(f"\u2551 {pad(name, width_p)}\u2502 {pad(value, width_v)}\u2502 {pad(unit, width_u)}\u2551")
            else:
                _result.append(f"{pad(name, width_p)} {pad(value, width_v)} {pad(unit, width_u)}")

        # print footer
        if self.draw_lines:
            _result.append("\u255a" + ("\u2550" * (width_p + 1)) + "\u2567" + ("\u2550" * (width_v + 1)) + "\u2567" + ("\u2550" * (width_u + 1)) + "\u255d")
        # _result.append("\n")
        return _result



import logging
import re

from .baseoutput import baseoutput
from ..helpers import get_kwargs
from ..helpers import key_wanted

log = logging.getLogger("mqtt")


class mqtt(baseoutput):
    def __str__(self):
        return "outputs the results to the supplied mqtt broker: eg {tag}/status/total_output_active_power/value 1250"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")

    def build_msgs(self, *args, **kwargs):

        data = get_kwargs(kwargs, "data")
        # Clean data
        command = data.pop("_command", None)
        data.pop("_command_description", None)
        data.pop("raw_response", None)

        # check if config supplied
        config = get_kwargs(kwargs, "config")
        if config is not None:
            log.debug(f"config: {config}")
            # get results topic
            results_topic = config.get("results_topic", None)
            # get formatting info
            remove_spaces = config.get("remove_spaces", True)
            keep_case = config.get("keep_case", False)
            filter = config.get("filter", None)
            excl_filter = config.get("excl_filter", None)
            tag = config.get("tag", None)
        else:
            results_topic = None
            # get formatting info
            remove_spaces = True
            keep_case = get_kwargs(kwargs, "keep_case")
            filter = get_kwargs(kwargs, "filter")
            excl_filter = get_kwargs(kwargs, "excl_filter")
            tag = get_kwargs(kwargs, "tag")

        if filter is not None:
            filter = re.compile(filter)
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)
        if tag is None:
            tag = command

        # build topic prefix
        if results_topic is not None:
            topic_prefix = results_topic
        else:
            topic_prefix = f"{tag}/status"

        # build data to output
        _data = {}
        for key, _values in data.items():
            # remove spaces
            if remove_spaces:
                key = key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, filter, excl_filter):
                _data[key] = _values
        log.debug(f"output data: {_data}")

        # Build array of mqtt messages
        msgs = []
        # Loop through responses
        for key, values in _data.items():
            value = values[0]
            unit = values[1]
            log.debug(
                f"build_msgs: prefix {topic_prefix}, key {key}, value {value}, unit {unit}"
            )
            msg = {"topic": f"{topic_prefix}/{key}/value", "payload": value}
            msgs.append(msg)
            if unit:
                msg = {"topic": f"{topic_prefix}/{key}/unit", "payload": unit}
                msgs.append(msg)
        log.debug(f"build_msgs: {msgs}")
        return msgs

    def output(self, *args, **kwargs):
        log.info("Using output processor: mqtt")
        log.debug(f"kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        # exit if no data
        if data is None:
            return

        # get the broker instance
        mqtt_broker = get_kwargs(kwargs, "mqtt_broker")
        # exit if no broker
        if mqtt_broker is None:
            return

        # build the messages...
        msgs = self.build_msgs(**kwargs)
        log.debug(f"mqtt.output msgs {msgs}")

        # publish
        mqtt_broker.publishMultiple(msgs)



""" abstractport.py """
import logging
from abc import ABC, abstractmethod

from powermon.commands.result import Result
from powermon.protocols.abstractprotocol import AbstractProtocol
from powermon.commands.command import Command


log = logging.getLogger("Port")


class AbstractPort(ABC):
    """
    model for all ports
    """

    @classmethod
    @abstractmethod
    def from_config(cls, config=None):
        """ build port object from config dict """
        raise NotImplementedError

    def __init__(self, protocol: AbstractProtocol):
        self.protocol: AbstractProtocol = protocol
        self.error_message = None

    def connect(self) -> bool:
        """ default port connect function """
        log.debug("Port connect not implemented")
        return False

    def disconnect(self) -> None:
        """ default port disconnect function """
        log.debug("Port disconnect not implemented")

    @abstractmethod
    def is_connected(self) -> bool:
        """ default is_connected function """
        raise NotImplementedError

    @abstractmethod
    def send_and_receive(self, command: Command) -> Result:
        """ main worker function for port objects, specific to each port type """
        raise NotImplementedError

    @abstractmethod
    def to_dto(self):
        """ convert port object to data transfer object """
        raise NotImplementedError

    @property
    def protocol(self) -> AbstractProtocol:
        """ return the protocol associated with this port """
        return self._protocol

    @protocol.setter
    def protocol(self, value):
        log.debug("Setting protocol to: %s", value)
        self._protocol = value

    def run_command(self, command: Command) -> Result:
        """ run_command takes a command object, runs the command and returns a result object (replaces process_command) """
        log.debug("Command %s", command)

        # open port if it is closed
        if not self.is_connected():
            if not self.connect():
                raise ConnectionError(f"Unable to connect to port: {self.error_message}")
        # FIXME: what if still not connected....
        # should, log an error and wait to try to reconnect (increasing backoff times)

        # update run times
        command.touch()
        # update full_command - expand any template / add crc etc
        # updates every run incase something has changed
        command.full_command = self.protocol.get_full_command(command.code)

        # run the command via the 'send_and_receive port function
        result = self.send_and_receive(command)
        log.debug("after send_and_receive: %s", result)
        return result
